import tkinter as tk
from tkinter import ttk
from PIL import Image, ImageTk
from tkinter import filedialog
from tkinter import messagebox
import os
from math import sqrt

class AutoScrollbar(ttk.Scrollbar):
    def set(self, lo, hi):
        if float(lo) <= 0.0 and float(hi) >= 1.0:
            self.grid_remove()
        else:
            self.grid()
            ttk.Scrollbar.set(self, lo, hi)

    def pack(self, **kw):
        raise tk.TclError('Cannot use pack with this widget')

    def place(self, **kw):
        raise tk.TclError('Cannot use place with this widget')

class Tool(ttk.Frame):
    def __init__(self, mainframe):
        # init with respect to root tk element
        ttk.Frame.__init__(self, master=mainframe)
        self.master.title('Tool') # title of the window

        # creating the image frame, the window part where the image will be displayed
        self.imageF = tk.Frame(self.master)
        self.imageF.grid(row=0, column=2, sticky='nswe') # placing the image frame to the right of the app window

        # vertical and horizontal scrollbars for canvas
        vbar = AutoScrollbar(self.imageF, orient='vertical') # reference for the vertical scrollbar
        hbar = AutoScrollbar(self.imageF, orient='horizontal') # reference for the horizontal scrollbar
        vbar.grid(row=0, column=1, sticky='ns') # placing the scrollbars in to the image frame
        hbar.grid(row=1, column=0, sticky='we')

        # Create canvas and put image on it
        self.canvas = tk.Canvas(self.imageF, highlightthickness=0,
                                xscrollcommand=hbar.set, yscrollcommand=vbar.set)
        self.canvas.grid(row=0, column=0, sticky='nswe')
        self.canvas.update()  # wait till canvas is created
        vbar.configure(command=self.scroll_y)  # bind scrollbars to the canvas
        hbar.configure(command=self.scroll_x)

        # Make the canvas expandable
        self.master.rowconfigure(0, weight=1)
        self.master.columnconfigure(2, weight=1)
        self.imageF.rowconfigure(0, weight=1)
        self.imageF.columnconfigure(0, weight=1)

        # an empty frame for an empty place between the options and the image
        self.emptyF = tk.Frame(self.master, width=5)
        self.emptyF.grid(row=0, column=1)

        # vars
        self.color = (255,0,0) # default - red, color tuple
        self.strcolor = "Red" # color string for writing to the file
        self.path = None # path of the image file
        self.dots = {} # dots in the canvas
        self.radius = 5 # radius of the dot
        self.bind_type = "Move" # initial bind_type
        self.imageid = -1 # for storing the cropped image

        # frame for the operations and such
        self.commandF = tk.Frame(self.master)
        self.commandF.grid(row=0, column=0, sticky='n') # placing it on the left of the window

        # child frame for the file options
        self.opF = tk.Frame(self.commandF)
        self.opF.grid(row=0, column=0)

        # label or text for file options
        self.buttonL = tk.Label(self.opF, text='File Operations')
        self.buttonL.grid(row=0, column=0)
        # open button reference, binded with open_image function
        self.open_button = tk.Button(self.opF, text="Open", command=self.open_image, width=16)
        self.open_button.grid(row=1, column=0)
        # save button reference, binded with save_image function
        self.save_button = tk.Button(self.opF, text="Save", command=self.save_image, width=16)
        self.save_button.grid(row=2, column=0)

        # empty frame between color and file operations
        self.emptyF0 = tk.Frame(self.commandF, height=32)
        self.emptyF0.grid(row=1, column=0)

        # child frame for the color operations
        self.colorF = tk.Frame(self.commandF)
        self.colorF.grid(row=2, column=0)

        # label or text for label options
        self.colorL = tk.Label(self.colorF, text='Color Operations')
        self.colorL.grid(row=0, column=0)
        # button for changing color to red
        self.red_button = tk.Button(self.colorF, text="Red", command=self.red, width=16)
        self.red_button.grid(row=1, column=0)
        # button for changing color to green
        self.green_button = tk.Button(self.colorF, text="Green", command=self.green, width=16)
        self.green_button.grid(row=2, column=0)
        # button for changing color to yellow
        self.yellow_button = tk.Button(self.colorF, text="Yellow", command=self.yellow, width=16)
        self.yellow_button.grid(row=3, column=0)
        # button for changing color to blue
        self.blue_button = tk.Button(self.colorF, text="Blue", command=self.blue, width=16)
        self.blue_button.grid(row=4, column=0)
        # button for changing color to cyan
        self.cyan_button = tk.Button(self.colorF, text="Cyan", command=self.cyan, width=16)
        self.cyan_button.grid(row=5, column=0)
        
        # empty frame between circle and color operations
        self.emptyF2 = tk.Frame(self.commandF, height=32)
        self.emptyF2.grid(row=3, column=0)

        # child frame for circle operations
        self.radF = tk.Frame(self.commandF)
        self.radF.grid(row=4, column=0)

        # text for circle operations
        self.radL = tk.Label(self.radF, text='Adjust Circle')
        self.radL.grid(row=0, column=0, columnspan=4)
        # button for decreasing the radius of circle for 3
        self.min3 = tk.Button(self.radF, text='-3', command=self.dec3)
        self.min3.grid(row=1, column=0)
        # button for decreasing the radius of circle for 1
        self.min1 = tk.Button(self.radF, text='-1', command=self.dec1)
        self.min1.grid(row=1, column=1)
        # button for increasing the radius of circle for 1
        self.plus1 = tk.Button(self.radF, text='+1', command=self.inc1)
        self.plus1.grid(row=1, column=2)
        # button for increasing the radius of circle for 3
        self.plus3 = tk.Button(self.radF, text='+3', command=self.inc3)
        self.plus3.grid(row=1, column=3)

        # empty frame between color and image operations
        self.emptyF3 = tk.Frame(self.commandF, height=32)
        self.emptyF3.grid(row=5, column=0)

        # child frame for image operations
        self.modeF = tk.Frame(self.commandF)
        self.modeF.grid(row=6, column=0)

        # text for image operations
        self.modeL = tk.Label(self.modeF, text='Change Mode')
        self.modeL.grid(row=0, column=0)
        # change mode to drawing
        self.draw_button = tk.Button(self.modeF, text="Draw", command=lambda:self.change_event('draw'), width=16)
        self.draw_button.grid(row=1, column=0)
        # change mode to erasing
        self.erase_button = tk.Button(self.modeF, text="Erase", command=lambda:self.change_event('erase'), width=16)
        self.erase_button.grid(row=2, column=0)
        # change mode to moving 
        self.move_button = tk.Button(self.modeF, text="Move", command=lambda:self.change_event('move'), width=16)
        self.move_button.grid(row=3, column=0)
        # change mode to zooming
        self.zoom_button = tk.Button(self.modeF, text='Zoom', command=lambda:self.change_event('zoom'), width=16)
        self.zoom_button.grid(row=4, column=0)

        # empty frame between image operations and information text
        self.emptyF4 = tk.Frame(self.commandF, height=32)
        self.emptyF4.grid(row=7, column=0)

        # text for showing the current image operation
        self.cmodeL = tk.Label(self.commandF, text="Current mode: {}".format(self.bind_type))
        self.cmodeL.grid(row=8, column=0)
        # text for showing the current color used
        self.ccolorL = tk.Label(self.commandF, text="Current color: {}".format(self.strcolor))
        self.ccolorL.grid(row=9, column=0)
        # text for showing the current radius of the circle
        self.cradL = tk.Label(self.commandF, text="Current radius: {:.2f}".format(self.radius))
        self.cradL.grid(row=10, column=0)

        # all these frame and button placements, grid functions, work with respect to their parent object
        # for example image frame was put column number 2 for it to be at the right side of the window
        # the first two was operations frame and an empty frame

        # bind events to the Canvas
        self.canvas.bind('<Configure>', self.show_image)  # <Configure> is for canvas resizing, show_image function updates image
        # show_image function will be called whenever a change happens on the image frame
        self.canvas.bind('<ButtonPress-1>', self.move_from) # initial image operation mode is moving, so pressing the image calls the move_from function
        self.canvas.bind('<B1-Motion>',     self.move_to) # after pressing motion calls move_to function

        self.image = None  # image reference
        self.width, self.height = -1, -1 # image width and height

        self.imscale = 1.0  # total image scale
        self.circscale = 1.0 # total circle radius scale

        self.delta = 1.3  # how much the image will change while using the zoom function
        self.circdelta = 1.3 # how much the circle radius will change while using the zoom function

        # put image into container rectangle and use it to set proper coordinates to the image
        # this container is for calculating the relative coordinates
        self.container = -1

    def scroll_y(self, *args, **kwargs):
        # scroll canvas vertically and redraw the image
        self.canvas.yview(*args, **kwargs)  # scroll vertically
        self.show_image()  # redraw the image

    def scroll_x(self, *args, **kwargs):
        # scroll canvas horizontally and redraw the image
        self.canvas.xview(*args, **kwargs)  # scroll horizontally
        self.show_image()  # redraw the image

    def move_from(self, event):
        # remember previous coordinates for scrolling with the mouse
        self.canvas.scan_mark(event.x, event.y)

    def move_to(self, event):
        # move canvas to the new position
        self.canvas.scan_dragto(event.x, event.y, gain=1)
        self.show_image()  # redraw the image

    def zoom(self, event):
        # main zoom function

        # these coordinates respect to the visible area of the canvas
        # suppose you are seeing an 50x50 part of a big image. top left of the visible area can be anywhere in the image
        # but when clicked it will be returned as (0, 0), since it is relative
        x = self.canvas.canvasx(event.x) # get the x coordinate of the mouse
        y = self.canvas.canvasy(event.y) # get the y coordinate of the mouse

        bbox = self.canvas.bbox(self.container)  # get image area relative to the canvas 

        if bbox[0] < x < bbox[2] and bbox[1] < y < bbox[3]: 
            pass  # check if the user clicking on the image
        else: 
            return  # do nothing if the click is not on the image
        
        scale = 1.0 
        circscale = 1.0
        # for event.num, 3 is right click for mouse. it is meant for zoom out function
        # in mac (for touchpad) this value is 2 
        if (event.num == 3):  # scroll down
            i = min(self.width, self.height)
            if (int(i * self.imscale) < 30): 
                return  # image is less than 30 pixels, cannot zoom out more
            
            self.imscale /= self.delta # updating the total scales applied to the image
            self.circscale /= self.circdelta

            circscale /= self.circdelta # updating the current scale used
            scale        /= self.delta
        
        # for event.num, 1 is right click for mouse. it is meant for zoom in function
        # in mac (for touchpad) this value is also 1 
        if (event.num == 1): 
            i = min(self.canvas.winfo_width(), self.canvas.winfo_height())
            if (i < self.imscale): 
                return  # 1 pixel is bigger than the visible area, cannot zoom in more
                
            self.imscale *= self.delta # updating the total scales applied to the image
            self.circscale *= self.circdelta

            circscale *= self.circdelta # updating the current scale used
            scale        *= self.delta
        
        self.radius *= circscale # update radius
        self.canvas.scale('all', x, y, scale, scale)  # rescale all canvas objects
        self.show_image() # redraw the canvas
        self.config_msg() # rewrite the information text

    def show_image(self, event=None):
        # rerender the image on the canvas

        # get image area relative to the canvas 
        bbox1 = self.canvas.bbox(self.container)  
        
        # Remove 1 pixel shift at the sides of the bbox1
        bbox1 = (bbox1[0] + 1, bbox1[1] + 1, bbox1[2] - 1, bbox1[3] - 1)

        # get visible area of the canvas
        # this is for when image moved such that there is a visible are that empty 
        bbox2 = (self.canvas.canvasx(0), 
                 self.canvas.canvasy(0),
                 self.canvas.canvasx(self.canvas.winfo_width()),
                 self.canvas.canvasy(self.canvas.winfo_height()))
        
        # get scroll region box
        # this means total visible canvas coordinates
        bbox = [min(bbox1[0], bbox2[0]), min(bbox1[1], bbox2[1]),  
                max(bbox1[2], bbox2[2]), max(bbox1[3], bbox2[3])]
        
        if bbox[0] == bbox2[0] and bbox[2] == bbox2[2]:  # whole image in the visible area
            bbox[0] = bbox1[0] # replace the image coordinates as the bounding box's coordinates
            bbox[2] = bbox1[2]
        if bbox[1] == bbox2[1] and bbox[3] == bbox2[3]:  # whole image in the visible area
            bbox[1] = bbox1[1]  # replace the image coordinates as the bounding box's coordinates
            bbox[3] = bbox1[3]
        
        # scroll only if the image left out the visible area
        self.canvas.configure(scrollregion=bbox)  # set scroll region

        x1 = max(bbox2[0] - bbox1[0], 0)  # get coordinates (x1,y1,x2,y2) of the image tile
        y1 = max(bbox2[1] - bbox1[1], 0)
        x2 = min(bbox2[2], bbox1[2]) - bbox1[0]
        y2 = min(bbox2[3], bbox1[3]) - bbox1[1]

        if (int(x2 - x1) > 0 and int(y2 - y1) > 0):  # show image if it in the visible area
            x = min(int(x2 / self.imscale), self.width)   # sometimes it is larger on 1 pixel
            y = min(int(y2 / self.imscale), self.height)  # and sometimes not

            # crop the image
            image = self.image.crop((int(x1 / self.imscale), int(y1 / self.imscale), x, y)) 

            # resize the image for showing it on the visible area
            imagetk = ImageTk.PhotoImage(image.resize((int(x2 - x1), int(y2 - y1))))

            # imageid reference is for accessing the image object on the canvas
            self.imageid = self.canvas.create_image(max(bbox2[0], bbox1[0]), max(bbox2[1], bbox1[1]),
                                               anchor='nw', image=imagetk)
            
            self.canvas.lower(self.imageid)  # set image into background, this is for circles
            self.canvas.imagetk = imagetk  # keep an extra reference to prevent garbage-collection

    def open_image(self, path=None):
        if (path is None):	# if the path was not passed as a parameter
            try:
                p = None

                if (self.path): # if some path is stored before, this is None in the first time application run
                    # split the filepath and the filename
                    p = os.path.split(self.path)[0] 

                # store the path
                path = filedialog.askopenfilename(initialdir=p, title='Open image', filetypes=[("Image files", ("*.png", "*.jpg", "*.jpeg")),("PNG files", "*.png"),("JPG files", "*.jpg"),("JPEG files", "*.jpeg"),("all files", "*.*")])
            
            # handle exceptions
            except(OSError, FileNotFoundError):
                messagebox.showerror('Error','Unable to find or open file <filename>')
            except Exception as error:
                messagebox.showerror('Error','An error occurred: <error>')

        if (path):
            self.path = path # store the path
            self.image = Image.open(path) # open the image
            self.width, self.height = self.image.size # store the image size informations
            self.canvas.delete("all") # delete all the current elements on the canvas
            self.dots.clear() # clear the dots dictionary since we are deleting them

            self.container = self.canvas.create_rectangle(0, 0, self.width, self.height, width=0) # create the image container

            self.imscale = 1.0 # reset the scales
            self.circscale = 1.0
            self.radius = 5

            self.show_image() # rerender the canvas
            self.config_msg() # rerender the information text

            try: # try loading the dot information from a csv
                self.load_image()
            except: # if not do nothing
                pass

    # next 5 functions is for changing color information
    # this is dirty coding for this event
    # it can be done waay better than this
    def red(self, event=None):
        self.color = (255, 0, 0)
        self.strcolor = "Red"
        self.config_msg()
    
    def green(self, event=None):
        self.color = (0, 255, 0)
        self.strcolor = "Green"
        self.config_msg()

    def yellow(self, event=None):
        self.color = (255, 255, 0)
        self.strcolor = "Yellow"
        self.config_msg()

    def blue(self, event=None):
        self.color = (0, 0, 255)
        self.strcolor = "Blue"
        self.config_msg()

    def cyan(self, event=None):
        self.color = (0, 255, 255)
        self.strcolor = "Cyan"
        self.config_msg()

    # next 4 functions is for changing circle radius
    # this is dirty coding for this event
    # it can be done waay better than this 
    def inc1(self):
        self.radius += 1

        self.config_msg()

    def inc3(self):
        self.radius += 3

        self.config_msg()

    def dec1(self):
        if(self.radius > 1):
            self.radius -= 1

            self.config_msg()
    
    def dec3(self):
        if(self.radius > 1):
            if(self.radius == 2 or self.radius == 3):
                self.radius = 1
            else:
                self.radius -= 3
            
            self.config_msg()

    # this function is for changing the image operation
    def change_event(self, t):
        if(t == 'draw'): # if the passed event is the draw event
            self.bind_type = "Draw" # store the event type for information text
            self.canvas.unbind('<ButtonPress-1>') # unbind the left mouse button events
            self.canvas.unbind('<B1-Motion>')
            self.canvas.bind('<Button-1>', self.draw_circle) # bind the draw_circle event to left mouse button
        # the rest checks are for different operations
        elif(t == 'erase'):
            self.bind_type = "Erase"
            self.canvas.unbind('<ButtonPress-1>')
            self.canvas.unbind('<B1-Motion>')
            self.canvas.bind('<Button-1>', self.erase_circle)
        elif(t == 'move'):
            self.bind_type = "Move"
            self.canvas.unbind('<Button-1>')
            self.canvas.bind('<ButtonPress-1>', self.move_from)
            self.canvas.bind('<B1-Motion>',     self.move_to)
        elif(t == 'zoom'):
            self.bind_type = "Zoom"
            self.canvas.unbind('<ButtonPress-1>')
            self.canvas.unbind('<B1-Motion>')
            self.canvas.bind('<Button-1>', self.zoom)
            self.canvas.bind('<Button-3>', self.zoom)
            # i am not sure about this button-3. in windows and linux it is working. if i remember correctly
            # it is the same in mac. one can try replacing it with the button-2 to see if that works

        self.config_msg()

    # this function is for drawing circles on the canvas
    # it also stores the coordinates of the centers of the circles with respect to image itself
    def draw_circle(self, event):
        r = self.radius # get radius

        bbox = (self.canvas.canvasx(0),  # get visible area of the canvas with respect to canvas itself
                self.canvas.canvasy(0),
                self.canvas.canvasx(self.canvas.winfo_width()),
                self.canvas.canvasy(self.canvas.winfo_height()))
        
        bbox1 = self.canvas.bbox(self.container)  # get image area with respect to canvas

        # event.x is as i told before respects to the visible area, if the mouse is at the top left, event.x is 0
        # since this is the case we need to find coordinates with respect to canvas to draw
        x = bbox[0] + event.x 
        y = bbox[1] + event.y

        # create_oval function returns an id for the created object, so that we can manipulate it when the time comes
        # and i tag these objects with a string, "circle"
        circle_id = self.canvas.create_oval(x-r, y-r, x+r, y+r, fill="#%02x%02x%02x" % self.color, outline="#%02x%02x%02x" % self.color, tags=('circle'))

        # now is the time to get circle coordinates with respect to image itself
        # because this is our main objective
        c = bbox[0] - bbox1[0] + event.x # calculate the x 
        d = bbox[1] - bbox1[1] + event.y # calculate the y
        w = (bbox1[2] - bbox1[0])/self.width # this calculations are for scaling purposes, since we want to coordinates with respect to the original image sizes
        h = (bbox1[3] - bbox1[1])/self.height # they are basically done for finding how much the image is strechted

        # store the circle itself in the dictionary with its coordinates with respect to image
        self.dots[circle_id] = (str(c/w), str(d/h), self.strcolor.lower())

    # this function is for writing the coordinate information on an .csv file
    # this .csv file's name is the same as image's name
    def save_image(self):

        if (self.path and len(self.dots) > 0):
            index = self.path.rfind(".") # find the extension dot

            # open the .csv file in the write mode and save it where the image at
            with open(self.path[:index] + ".csv", "w") as f:
                for key, value in self.dots.items():
                    f.write(",".join(value) + '\n')
            
            # inform the user that coordinates are now saved
            messagebox.showinfo("Saved", "Coordinates are saved.")

            self.config_msg()

    # this function is for loading purposes
    # that is the desired .csv is provided
    # it searches for <image_name>.csv file and loads the circles on to the canvas
    def load_image(self):
        try:
            # find the extension dot
            index = self.path.rfind(".")

            # this time open .csv file in read mode
            with open(self.path[:index] + ".csv", "r") as f:
                # in every line of the .csv file
                for line in f:
                    # seperate the values
                    line = line.split(',')

                    # store the values
                    x, y, color = line[0], line[1], line[2][:-1] # in the color variable discard the '\n'

                    # x and y are the coordinates
                    x = float(x)
                    y = float(y)

                    # a variable for storing the hex value of the color
                    hexcolor = None
                    
                    # set the hexcolor variable
                    if(color == "red"):
                        hexcolor = "#ff0000"
                    elif(color == "green"):
                        hexcolor = "#00ff00"
                    elif(color == "yellow"):
                        hexcolor = "#ffff00"
                    elif(color == "blue"):
                        hexcolor = "#0000ff"
                    elif(color == "cyan"):
                        hexcolor = "#00ffff"

                    # store the circles in the dictionary
                    # the coordinate operations are not needed here since, at first
                    # the image and the canvas are the same size
                    circle_id = self.canvas.create_oval(x-4, y-4, x+4, y+4, fill=hexcolor, outline=hexcolor, tags=('circle'))
                    self.dots[circle_id] = (str(x), str(y), color)

            # inform the user that the circles are loaded
            messagebox.showinfo("Loaded", "Dots are loaded.")
        
        # if the required file is not found
        except Exception as e:
            messagebox.showerror('Error','No saves found.')

    # this function is for the erase operation
    def erase_circle(self, event):

        bbox = (self.canvas.canvasx(0),  # get visible area of the canvas 
                self.canvas.canvasy(0),
                self.canvas.canvasx(self.canvas.winfo_width()),
                self.canvas.canvasy(self.canvas.winfo_height()))

        x = bbox[0] + event.x # get the mouse coordinates with respect to canvas
        y = bbox[1] + event.y

        # this is find_closest function of the canvas
        # it searches the canvas to find the closest object
        # this check is for clicking the circle itself
        # i believed the canvas' function would be faster and did this check
        item = self.canvas.find_closest(x, y)

        # if the clicked object is a circle delete directly
        if("circle" in self.canvas.gettags(item)):
            self.canvas.delete(item[0])
            self.dots.pop(item[0])
        else: # call the closest function to find the closest circle this time and delete it
            circle_id = self.closest(event)

            if(circle_id == -1):
                return
            
            self.canvas.delete(circle_id)
            self.dots.pop(circle_id)

    # this function is for configuring the information text
    def config_msg(self):
        # i call configure function on all the label objects created at __init__ function
        self.cmodeL.configure(text="Current mode: {}".format(self.bind_type))
        self.ccolorL.configure(text="Current color: {}".format(self.strcolor))
        self.cradL.configure(text="Current radius: {:.2f}".format(self.radius))

    # this is a simple function for finding the closest circle in the canvas with respect to clicked position
    # it uses euclidian distances
    def closest(self, event):
        r = self.radius # get radius

        bbox = (self.canvas.canvasx(0),  # get visible area of the canvas
                self.canvas.canvasy(0),
                self.canvas.canvasx(self.canvas.winfo_width()),
                self.canvas.canvasy(self.canvas.winfo_height()))

        bbox1 = self.canvas.bbox(self.container)  # get image area

        # calculate the needed variables
        c = bbox[0] - bbox1[0] + event.x
        d = bbox[1] - bbox1[1] + event.y
        w = (bbox1[2] - bbox1[0])/self.width
        h = (bbox1[3] - bbox1[1])/self.height

        # find the coordinates with respect to image
        x = c/w
        y = d/h

        # define a variable for return value
        closest_id = -1
        # define a variable for smallest distance
        distance = float('inf')

        # simply loop over the dots
        for key, value in self.dots.items():
            # get the coordinates of the dots
            vx, vy = float(value[0]), float(value[1])

            # calculate the differences between coordinates
            diffx, diffy = vx-x, vy-y

            # calculate the total distance from mouse coordinate to dot's coordinate
            dist = sqrt(diffx*diffx + diffy*diffy)

            # if the calculate distance is a smaller one store it
            if dist < distance:
                closest_id = key
                distance = dist

        return closest_id

# create a Tk window
root = tk.Tk()

# create Tool object with root as its master
app = Tool(root)

# run the root's mainloop
root.mainloop()